/***********************************************************************/
/*                                                                     */
/* infoman.cpp : Source file for SP_FILEINFO cache class               */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/infoman.cpp $                */
/*  $Date: 2005-11-14 20:30:56 +0900 (Mon, 14 Nov 2005) $              */
/*   $Rev: 54 $                                                        */
/*                                                                     */
/***********************************************************************/

#include <iostream>
#include <algorithm>

#include "odstream.hpp"
#include "infoman.hpp"

using yak::utils::windows::PIDL;
using yak::debug::ods;

InfoManager::InfoManager(int nMaxNum_):nMaxNum(nMaxNum_)
{
	ods << "InfoManager::InfoManager(" << nMaxNum_ << ')' << std::endl;
	/* TODO: check if nMaxNum is positive value */
}

InfoManager::~InfoManager()
{
	ods << "InfoManager::~InfoManager()" << std::endl;
}

std::size_t InfoManager::getFileInfoSize(const KeyType& kFilename) const
{
	ods << "InfoManager::getFileInfoSize(" << kFilename << ')' << std::endl;
	if(!isCached(kFilename)) {
		// TODO: should throw
		return 0;
	}
	return getFileInfoCount(kFilename) * sizeof(SPI_FILEINFO);
}

std::size_t InfoManager::copyFileInfo(const KeyType &kFilename, void* pInfo, std::size_t sSize) const
{
	ods << "InfoManager::copyFileInfo(" << kFilename << ',' << pInfo << ',' << sSize << ')' << std::endl;
	if(!isCached(kFilename)) {
		// TODO: should throw
		return 0;
	}
	std::size_t sCopySize = getFileInfoSize(kFilename);
	if(sSize < sCopySize) {
		// TODO: should throw
		return 0;
	}

	CopyMemory(pInfo, &(mTable.find(kFilename)->second.first[0]), sCopySize);
	// TODO: check error?
	adjustLRU(kFilename);
	return sCopySize;
}

std::size_t InfoManager::getFileInfoCount(const KeyType &kFilename) const
{
	ods << "InfoManager::getFileInfoCount(" << kFilename << ')' << std::endl;
	if(!isCached(kFilename)) {
		// TODO: should throw
		return 0;
	}
	return mTable.find(kFilename)->second.first.size();
}

struct PathChecker
{
	PathChecker(const std::string& sPath_):sPath(sPath_) {}
	bool operator()(const SPI_FILEINFO& spInfo) {
		if(std::string(spInfo.path) + spInfo.filename == sPath) return true;
		return false;
	}
	const std::string& sPath;
};

void InfoManager::getFileInfo(const KeyType &kFilename, const std::string& sPath, SPI_FILEINFO* pResult) const
{
	ods << "InfoManager::getFileInfo(" << kFilename << ',' << sPath << ',' << pResult << ')' << std::endl;
	if(!isCached(kFilename)) {
		ods << "InfoManager::getFileInfo() - not cached!!!" << std::endl;
		// TODO: should throw proper exception
		throw 1;
	}
	const std::vector<SPI_FILEINFO> &vFileInfo = mTable.find(kFilename)->second.first;
	std::vector<SPI_FILEINFO>::const_iterator ci = std::find_if(vFileInfo.begin(), vFileInfo.end(), PathChecker(sPath));
	if(ci != vFileInfo.end()) {
		*pResult = *ci;
	} else {
		ods << "InfoManager::getFileInfo() - not found" << std::endl;
		// TODO: should throw proper exception
		throw 1;
	}
}

const PIDL& InfoManager::getPIDL(const KeyType &kFilename, DWORD dwPosition) const
{
	ods << "InfoManager::getPIDL(" << kFilename << ',' << dwPosition << ')' << std::endl;
	if(!isCached(kFilename)) {
		// TODO: should throw proper exception
		throw 1;
	}
	if(dwPosition >= getFileInfoCount(kFilename)) {
		// TODO: should throw proper exception
		throw 1;
	}
	adjustLRU(kFilename);
	return mTable.find(kFilename)->second.second[dwPosition];
}

void InfoManager::addFileInfo(const KeyType &kFilename, const SPI_FILEINFO *pFileInfo, const PIDL *pPIDL, int nNum)
{
	ods << "InfoManager::addFileInfo(" << kFilename << ',' << pFileInfo << ',' << pPIDL << ',' << nNum << ')' << std::endl;
	if(isCached(kFilename)) {
		addFileInfoImp(mTable[kFilename].first, mTable[kFilename].second, pFileInfo, pPIDL, nNum);
	} else { // New entry
		if(getFileInfoCount(kFilename) == static_cast<std::size_t>(nMaxNum)) { // Full
			mTable.erase(lLRU.back());
			lLRU.pop_back();
		}
		lLRU.push_front(kFilename);
		std::pair<TableType::iterator, bool> tti = mTable.insert(std::make_pair(kFilename, ValueType()));
		addFileInfoImp(tti.first->second.first, tti.first->second.second, pFileInfo, pPIDL, nNum);
	}
	adjustLRU(kFilename);
}

void InfoManager::flushAllFileInfo()
{
	ods << "InfoManager::flushAllFileInfo()" << std::endl;
	mTable.clear();
	lLRU.clear();
}

void InfoManager::flushFileInfo(const KeyType &kFilename)
{
	ods << "InfoManager::flushFileInfo(" << kFilename << ')' << std::endl;
	if(!isCached(kFilename)) {
		// TODO: should throw
		return;
	}
	mTable.erase(mTable.find(kFilename));
	std::list<KeyType>::iterator li = std::find(lLRU.begin(), lLRU.end(), kFilename);
	if(li == lLRU.end()) {
		// TODO: should throw
		return;
	}
	lLRU.erase(li);
}

void InfoManager::adjustLRU(const KeyType &kFilename) const
{
	ods << "InfoManager::adjustLRU(" << kFilename << ')' << std::endl;
	if(!isCached(kFilename)) {
		// TODO: should throw
		return;
	}
	if(lLRU.front() == kFilename) return;
	std::list<KeyType>::iterator li = find(lLRU.begin(), lLRU.end(), kFilename);
	if(li == lLRU.end()) {
		// TODO: should throw
		return;
	}
	lLRU.splice(lLRU.begin(), lLRU, li);
}

void InfoManager::addFileInfoImp(std::vector<SPI_FILEINFO>& vFileInfo, std::vector<PIDL>& vPIDL, const SPI_FILEINFO *pFileInfo, const PIDL *pPIDL, int nNum)
{
	ods << "InfoManager::addFileInfoImp(" << &vFileInfo << ',' << &vPIDL << ',' << pFileInfo << ',' << pPIDL << ')' << std::endl;
	vFileInfo.reserve(vFileInfo.size() + nNum);
	vPIDL.reserve(vPIDL.size() + nNum);
	vFileInfo.insert(vFileInfo.end(), pFileInfo, pFileInfo + nNum);
	vPIDL.insert(vPIDL.end(), pPIDL, pPIDL + nNum);
}
