/***********************************************************************/
/*                                                                     */
/* infoman.hpp : Header file for SP_FILEINFO cache class               */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/infoman.hpp $                */
/*  $Date: 2005-11-14 20:30:56 +0900 (Mon, 14 Nov 2005) $              */
/*   $Rev: 54 $                                                        */
/*                                                                     */
/***********************************************************************/

#ifndef INFOMAN_HPP
#define INFOMAN_HPP

#include <string>
#include <vector>
#include <map>
#include <list>

#include "Spi_api.h"
#include "PIDL.hpp"

// TODO: Make thread-safe
// TODO: Contents-based cache is preferable, probably.
// TODO: Time-based flush
class InfoManager
{
public:
	typedef std::string KeyType;
	InfoManager(int nMaxNum_);
	~InfoManager();
	bool isCached(const KeyType& kFilename) const {
		return mTable.find(kFilename) != mTable.end();
	}
	std::size_t getFileInfoSize(const KeyType& kFilename) const;
	std::size_t copyFileInfo(const KeyType &kFilename, void* pInfo, std::size_t sSize) const;
	std::size_t getFileInfoCount(const KeyType &kFilename) const;
	void getFileInfo(const KeyType &kFilename, const std::string& sPath, SPI_FILEINFO* pResult) const;
	// Return reference to an entity in this object.
	const yak::utils::windows::PIDL& getPIDL(const KeyType &kFilename, DWORD dwPosition) const;
	// position is adjusted to index of list in this object.
	// PIDL is copied deeply.
	void addFileInfo(const KeyType &kFilename, const SPI_FILEINFO *pFileInfo, const yak::utils::windows::PIDL *pPIDL, int nNum);
	void flushAllFileInfo();
	void flushFileInfo(const KeyType &kFilename);
private:
	void adjustLRU(const KeyType &kFilename) const;
	void addFileInfoImp(std::vector<SPI_FILEINFO>& vFileInfo, std::vector<yak::utils::windows::PIDL>& vPIDL, const SPI_FILEINFO *pFileInfo, const yak::utils::windows::PIDL *pPIDL, int nNum);
	typedef std::pair<std::vector<SPI_FILEINFO>, std::vector<yak::utils::windows::PIDL> > ValueType;
	typedef std::map<KeyType, ValueType> TableType;
	std::map<KeyType, ValueType> mTable;
	mutable std::list<KeyType> lLRU;
	int nMaxNum;
};

#endif
