/***********************************************************************/
/*                                                                     */
/* spdir.cpp   : Source file for handling special folder               */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/spdir.cpp $                  */
/*  $Date: 2005-11-14 23:01:01 +0900 (Mon, 14 Nov 2005) $              */
/*   $Rev: 56 $                                                        */
/*                                                                     */
/***********************************************************************/

#include <iostream>
#include <string>

#include <shlobj.h>
#include <shlwapi.h>

#include "PIDL.hpp"
#include "Slurp.hpp"
#include "odstream.hpp"
#include "infoman.hpp"

using yak::utils::windows::PIDL;
using yak::utils::windows::Slurp;
using yak::debug::ods;

using namespace std;

// TODO: Use singleton (or separate action, for example, to use iterator interface)
extern InfoManager g_imSpdir;

// from afxspdir.cpp
extern int safe_strncpy(char *dest, const char *src, std::size_t n);

// TODO: Make OO interface

static IShellFolder* GetShellFolder(LPOLESTR pCLSID, PIDL &pidl)
{
	HRESULT hRes;
	IShellFolder *pDesktop, *pResult;
	ULONG uLen;

	hRes = SHGetDesktopFolder(&pDesktop);
	if(hRes != S_OK) {
		ods << "ERROR on SHGetDesktopFolder" << endl;
		return NULL;
	}
	if(pCLSID == NULL) return pDesktop;
	hRes = pDesktop->ParseDisplayName(NULL, NULL, pCLSID, &uLen, pidl, NULL);
	if(hRes != S_OK) {
		ods << "ERROR on ParseDisplayName" << endl;
		pDesktop->Release();
		return NULL;
	}
	if(pidl) {
		ods << pidl << endl;
	}
	hRes = pDesktop->BindToObject(pidl, NULL, IID_IShellFolder, (LPVOID*)&pResult);
	pDesktop->Release();
	if(hRes == S_OK) return pResult;
	return NULL;
}

// from SDK example HRESULT CreateLink(LPCSTR lpszPathObj, LPCSTR lpszPathLink, LPCSTR lpszDesc)
static HRESULT CreateLink(LPITEMIDLIST pidl, LPCSTR lpszPathLink, LPCSTR lpszDesc) 
{ 
    HRESULT hres; 
    IShellLink* psl; 

	ods << "CreateLink(," << lpszPathLink << ',' << lpszDesc << ')' << std::endl;

    // Get a pointer to the IShellLink interface. 
    hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, 
                            IID_IShellLink, (LPVOID*)&psl); 
    if (SUCCEEDED(hres)) 
    { 
		ods << "CreateLink() : CoCreateInstance() suceeds," << std::endl;
 
        IPersistFile* ppf; 
 
        // Set the path to the shortcut target and add the description. 
        hres = psl->SetIDList(pidl); 
        if (!SUCCEEDED(hres)){
			ods << "SetIDList is failed" << endl;
		} 
        psl->SetDescription(lpszDesc); 
 
        // Query IShellLink for the IPersistFile interface for saving the 
        // shortcut in persistent storage. 
        hres = psl->QueryInterface(IID_IPersistFile, (LPVOID*)&ppf); 
 
        if (SUCCEEDED(hres)) 
        { 
            WCHAR wsz[MAX_PATH]; 
 
            // Ensure that the string is Unicode. 
            MultiByteToWideChar(CP_ACP, 0, lpszPathLink, -1, wsz, MAX_PATH); 
			
            // TODO: Check return value from MultiByteWideChar to ensure success.
 
            // Save the link by calling IPersistFile::Save. 
            hres = ppf->Save(wsz, TRUE); 
	        if (SUCCEEDED(hres)){
				ods << "IPersistFile->Save() succeeds" << endl;
			} else {
				ods << "IPersistFile->Save() is failed" << endl;
			} 
            ppf->Release(); 
        } else {
			ods << "QueryInterface for IPersistFile is failed" << endl;
		}
        psl->Release(); 
    } 
    return hres; 
}

HANDLE CreateLinkOnMemory(LPITEMIDLIST pidl, LPSTR lpszDesc)
{
	char szTempPath[MAX_PATH], szTempName[MAX_PATH];
	HANDLE hResult = NULL;

	GetTempPath(sizeof(szTempPath), szTempPath);
	GetTempFileName(szTempPath, "asd", 0, szTempName);

	CreateLink(pidl, szTempName, lpszDesc);

	{ // Restrict lifetime of Slurp for DeleteFile
		Slurp<BYTE> sLink(szTempName);
		if(sLink.getSize() == 0) { // Probably, CreateLink failed actually. TODO: should be detected by exception
			char szPath[MAX_PATH];
			if(SHGetPathFromIDList(pidl, szPath)) {
				Slurp<BYTE> sLink(szPath);
				hResult = LocalAlloc(LMEM_MOVEABLE, sLink.size());
				LPVOID pv = LocalLock(hResult);
				CopyMemory(pv, sLink.begin(), sLink.size());
				LocalUnlock(hResult);
			}
		} else {
			hResult = LocalAlloc(LMEM_MOVEABLE, sLink.size());
			LPVOID pv = LocalLock(hResult);
			CopyMemory(pv, sLink.begin(), sLink.size());
			LocalUnlock(hResult);
		}
	}

	DeleteFile(szTempName);

	return hResult;
}

static void MakeFSSafe(std::string &sTarget)
{
	extern char cConvert[256]; // from table.cpp
	for(std::string::iterator i = sTarget.begin(); i != sTarget.end(); ++i) {
		char c = cConvert[static_cast<unsigned char>(*i)];
		if(c == 0) {
			if(++i == sTarget.end()) break;
		} else {
			*i = c;
		}
	}
}

static void DumpTree(IShellFolder *pShellFolder, const PIDL& pidl_base, const std::string &sKey, int &entry, int depth = 0, int max_entry = 0, int max_entry_indir = 0, const string& base = "")
{
	PIDL pidl;
	STRRET sr;
	LPTSTR pszName;
	IEnumIDList *pEnum;
	IShellFolder *pSubFolder;
	SPI_FILEINFO spfi;
	int nEntryInDir = 0;
	std::string sDispName;

	ods << "DumpTree(" << pShellFolder << ',' << ',' << sKey << ',' << entry << ',' << depth << ',' << max_entry << ',' << max_entry_indir << ',' << base << ')' << std::endl;

	if(!pShellFolder) return;

	if(max_entry != 0 && entry == max_entry) return;

	std::string sPrefix = (base == "") ? base : base + "\\";

	HRESULT hRes = pShellFolder->EnumObjects(NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS | SHCONTF_INCLUDEHIDDEN, &pEnum);
	if(hRes == S_OK && pEnum != NULL) {
		hRes = pEnum->Next(1, pidl, NULL);
		while(hRes == NOERROR) {
			pShellFolder->GetDisplayNameOf(pidl, SHGDN_NORMAL, &sr);
			hRes = StrRetToStr(&sr, pidl, &pszName);
			if(hRes == S_OK) {
				if((max_entry == 0 || entry < max_entry) && (max_entry_indir == 0 || nEntryInDir < max_entry_indir)) {
					ZeroMemory(&spfi, sizeof(spfi));
					safe_strncpy(reinterpret_cast<char*>(spfi.method), "SPDIR", sizeof(spfi.method));
					spfi.position = entry++;
					++nEntryInDir;
					MakeFSSafe(sDispName = pszName);
					std::string sName = sDispName + ".lnk";
					safe_strncpy(spfi.filename, sName.c_str(), sizeof(spfi.filename));
					if(base != "") {
						safe_strncpy(spfi.path, sPrefix.c_str(), sizeof(spfi.path));
					}
					ods << "sKey: " << sKey << " path: " << spfi.path << " filename: " << spfi.filename << std::endl;
					const PIDL& pidl_ = pidl_base + pidl;
					g_imSpdir.addFileInfo(sKey, &spfi, &pidl_, 1);
				}
			}
//			ulAttr = SFGAO_BROWSABLE;
//			pShellFolder->GetAttributesOf(1, pidl, &ulAttr);
//			if(ulAttr & SFGAO_BROWSABLE && depth != 1) {
			if(depth != 1) {
				hRes = pShellFolder->BindToObject(pidl, NULL, IID_IShellFolder, (LPVOID*)&pSubFolder);
				if(hRes == S_OK && pSubFolder) {
					DumpTree(pSubFolder, pidl_base + pidl, sKey, entry, depth == 0 ? 0 : depth - 1, max_entry, max_entry_indir, sPrefix + sDispName);
					pSubFolder->Release();
				}
			}
			CoTaskMemFree(pszName);
			if((max_entry != 0 && entry == max_entry) || (max_entry_indir != 0 && nEntryInDir == max_entry_indir)) break;
			hRes = pEnum->Next(1, pidl, NULL);
		}
		pEnum->Release();
	}
}

void MakeFileInfo(LPOLESTR pCLSID, const std::string &sKey, int depth, int max_entry, int max_entry_indir)
{
	PIDL pidl;
	int nIndex = 0;

	ods << "MakeFileInfo(," << sKey << ',' << depth << ',' << max_entry << ',' << max_entry_indir << ')' << std::endl;

	IShellFolder *pShellFolder = GetShellFolder(pCLSID, pidl);
	if(pShellFolder) {
		DumpTree(pShellFolder, pidl, sKey, nIndex, depth, max_entry, max_entry_indir);
		pShellFolder->Release();
	}
}
